"""
Many implementations of the dot product.
See `Cython documentation <http://docs.cython.org/en/latest/>`_.
"""
from libc.stdlib cimport calloc, free
from libc.string cimport memcpy
from libc.stdio cimport printf
from libc.math cimport NAN

import numpy
cimport numpy
cimport cython
numpy.import_array()

def pyfilter_dmax(va, mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Python inside cython.
    
    :param va: first vector
    :param mx: maximum
    """
    for i in range(va.shape[0]):
        if va[i] > mx:
            va[i] = mx


def filter_dmax_cython(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Simple cython.
    
    :param va: first vector
    :param mx: maximum
    """
    for i in range(va.shape[0]):
        if va[i] > mx:
            va[i] = mx


@cython.boundscheck(False)
@cython.wraparound(False)
def filter_dmax_cython_optim(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Simple cython with no bound checked, no wrap around.
    
    :param va: first vector
    :param mx: maximum
    """
    for i in range(va.shape[0]):
        if va[i] > mx:
            va[i] = mx


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _cyfilter_dmax(double[::1] va, double mx) nogil:
    """
    Replaces all value superior to *mx* by *mx*.
    
    :param va: first vector
    :param mx: maximum
    """
    for i in range(va.shape[0]):
        if va[i] > mx:
            va[i] = mx


@cython.boundscheck(False)
@cython.wraparound(False)
def cyfilter_dmax(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Wraps a C function implemented in cython.
    
    :param va: first vector
    :param mx: maximum
    """
    with nogil:
        _cyfilter_dmax(va, mx)


cdef extern from "experiment_cython_.h":
    cdef void filter_dmax(double *p1, cython.int size, double mx) nogil
    cdef void filter_dmax2(double *p1, cython.int size, double mx) nogil
    cdef void filter_dmax16(double *p1, cython.int size, double mx) nogil
    cdef void filter_dmax4(double *p1, cython.int size, double mx) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
def cfilter_dmax(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Wraps a C function implemented in C.
        
    :param va: first vector
    :param mx: maximum
    """
    with nogil:
        filter_dmax(&va[0], va.shape[0], mx)


@cython.boundscheck(False)
@cython.wraparound(False)
def cfilter_dmax2(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Wraps a C function implemented in C,
    but uses operator ``? :`` instead of keyword ``if``.
        
    :param va: first vector
    :param mx: maximum
    """
    with nogil:
        filter_dmax2(&va[0], va.shape[0], mx)


@cython.boundscheck(False)
@cython.wraparound(False)
def cfilter_dmax16(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Wraps a C function implemented in C,
    but uses operator ``? :`` instead of keyword ``if``.
    Goes 16 by 16.
        
    :param va: first vector
    :param mx: maximum
    """
    with nogil:
        filter_dmax16(&va[0], va.shape[0], mx)


@cython.boundscheck(False)
@cython.wraparound(False)
def cfilter_dmax4(double[::1] va, double mx):
    """
    Replaces all value superior to *mx* by *mx*.
    Wraps a C function implemented in C,
    but uses operator ``? :`` instead of keyword ``if``.
    Goes 4 by 4.
        
    :param va: first vector
    :param mx: maximum
    """
    with nogil:
        filter_dmax4(&va[0], va.shape[0], mx)



# coding: utf-8
"""
Common methods about training, predicting for :epkg:`SIR` models.
"""
import pprint
import numpy
from sympy import Symbol, diff as sympy_diff
from sympy.core.numbers import Zero
from ..optim import SGDOptimizer


class BaseSIRSklAPI:
    """
    Common methods about training, predicting for :epkg:`SIR` models.
    """

    def _check_fit_predict(self, X, y=None):
        if not isinstance(X, numpy.ndarray):
            raise TypeError("X must be a numpy array.")
        if len(X.shape) != 2:
            raise ValueError("X must be a matrix.")
        clq = self.quantity_names
        if len(clq) != X.shape[1]:
            raise ValueError(
                "Unexpected number of columns, got {}, expected {}.".format(
                    X.shape[1], len(clq)))
        sums = numpy.sum(X, axis=1)
        mi, ma = sums.min(), sums.max()
        df = abs(ma - mi)
        if df > abs(ma) * 1e-3:
            raise ValueError(
                "All rows must sum up to the same amount. Current "
                "range: [{}, {}].".format(mi, max))
        if y is not None:
            if not isinstance(y, numpy.ndarray):
                raise TypeError("y must be a numpy array.")
            if y.shape != X.shape:
                raise ValueError(
                    "Unexpected shape of y, got {}, expected {}.".format(
                        y.shape, X.shape))
        return clq, ma

    def fit(self, X, y, t=0, max_iter=100,
            learning_rate_init=0.1, lr_schedule='constant',
            momentum=0.9, power_t=0.5, verbose=False):
        """
        Fits a model :class:`BaseSIR <aftercovid.models._base_sir.BaseSIR>`.

        :param X: known values for every quantity at time *t*,
            every column is mapped to the list returned by
            :meth:`quantity_names <aftercovid.models._base_sir.quantity_names>`
        :param y: known derivative for every quantity at time *t*,
            comes in the same order as *X*
        :param t: implicit feature
        :param max_iter: number of iteration
        :param learning_rate_init: see :class:`SGDOptimizer
            <aftercovid.optim.SGDOptimizer>`
        :param lr_schedule: see :class:`SGDOptimizer
            <aftercovid.optim.SGDOptimizer>`
        :param momentum: see :class:`SGDOptimizer
            <aftercovid.optim.SGDOptimizer>`
        :param power_t: see :class:`SGDOptimizer
            <aftercovid.optim.SGDOptimizer>`
        :param verbose: see :class:`SGDOptimizer
            <aftercovid.optim.SGDOptimizer>`
        Both *X* and *y* have the same shape.

        The training needs two steps. The first one creates a training
        datasets. The second one estimates the coefficients by using
        a stochastic gradient descent (see :class:`SGDOptimizer
        <aftercovid.optim.SGDOptimizer>`).
        Let's use a SIDR model (see :class:`CovidSIR
        <aftercovid.models.CovidSIR>`).as an example.
        Let's denote the parameters as :math:`\\Omega`
        and :math:`Z_1=S`, ...,  :math:`Z_4=R`.
        The model is defined by
        :math:`\\frac{dZ_i}{dt} = f_i(\\Omega, Z)`
        where :math:`Z=(Z_1, ..., Z_4)`.
        *y* is used to compute the expected derivates
        :math:`\\frac{dZ_i}{dt}`. The loss function is defined as

        .. math::

            L(\\Omega,Z) = \\sum_{i=1}^4 \\left( f_i(\\Omega,Z) -
            \\frac{dZ_i}{dt}\\right)^2

        Then the gradient is:

        .. math::

            \\frac{\\partial L(\\Omega,Z)}{\\partial\\Omega} =
            2 \\sum_{i=1}^4 \\frac{\\partial f_i(\\Omega,Z)}{\\partial\\Omega}
            \\left( f_i(\\Omega,Z) - \\frac{dZ_i}{dt} \\right)

        A stochastic gradient descent takes care of the rest.
        """
        self._fit(
            X,
            y,
            t=0,
            max_iter=max_iter,
            learning_rate_init=learning_rate_init,
            lr_schedule=lr_schedule,
            momentum=momentum,
            power_t=power_t,
            verbose=verbose)
        return self

    def _losses_sympy(self):
        """
        Builds the loss functions using :epkg:`sympy`,
        one for each quantity.
        """
        clq = self.quantity_names
        res = []
        for name in clq:
            sym = Symbol('d' + name)
            eq = self._eq[name]
            res.append((eq - sym) ** 2)
        return res

    def _grads_sympy(self):
        """
        Builds the gradient for every parameter,
        exactly number of quantities multiplied by
        the number of parameters.
        """
        losses = self._losses_sympy()
        clq = self.quantity_names
        prn = self.param_names
        res = []
        for name, eq in zip(clq, losses):
            row = []
            for pn in prn:
                pns = Symbol(pn)
                df = sympy_diff(eq, pns)
                row.append(df)
            res.append(row)
        return res

    def predict(self, X, t=0):
        """
        Predicts the derivative at time *t*.

        :param X: known values for every quantity at time *t*,
            every column is mapped to the list returned by
            :meth:`quantity_names <aftercovid.models._base_sir.quantity_names>`
        :param t: implicit feature
        :return: predictive derivative
        """
        cache = self._eval_cache()
        clq, N = self._check_fit_predict(X)
        if N != self['N']:
            raise ValueError(
                "All rows must sum up to {} not {}.".format(self['N'], N))
        pos = {n: i for i, n in enumerate(clq)}
        pred = numpy.empty(X.shape, dtype=X.dtype)

        for i in range(0, X.shape[0]):
            svalues = {self._syms['t']: t + i}
            svalues.update(cache)
            for n, v in zip(clq, X[i, :]):
                svalues[self._syms[n]] = v

            for k, v in self._eq.items():
                ev = v.evalf(subs=svalues)
                pred[i, pos[k]] = ev

        return pred

    def _fit(self, X, y, t, max_iter,
             learning_rate_init, lr_schedule,
             momentum, power_t, verbose):
        '''
        See method :meth:`fit
        <aftercovid.models._base_sir_estimation.BaseSIRSklAPI.fit>`
        and :class:`SGDOptimizer <aftercovid.optim.SGDOptimizer>`.
        '''
        clq, N = self._check_fit_predict(X, y)
        pnames = self.param_names
        self['N'] = N

        symbol_clq = [Symbol(n) for n in clq]
        symbol_d_clq = [Symbol('d' + n) for n in clq]
        symbol_N = Symbol('N')
        symbol_params = [Symbol(n) for n in pnames]

        # loss and gradients functions
        losses = self._losses_sympy()
        grads = self._grads_sympy()

        def pformat(d):  # pragma: no cover
            nd = {str(k): (str(v), type(v), type(k)) for k, v in d.items()}
            return pprint.pformat(nd)

        def fct_loss(coef, X, y):
            'Computes the loss function for every X and y.'
            svalues = {symbol_N: N}
            for p, c in zip(symbol_params, coef):
                svalues[p] = c

            res = 0.
            for i in range(X.shape[0]):
                for n, v in zip(symbol_clq, X[i, :]):
                    svalues[n] = v
                for n, v in zip(symbol_d_clq, y[i, :]):
                    svalues[n] = v
                for loss in losses:
                    try:
                        res += loss.evalf(subs=svalues)
                    except (AttributeError, TypeError,  # pragma: no cover
                            IndexError) as e:
                        raise RuntimeError(
                            'Unable to calculate loss for [{}] with '
                            'values={}.'.format(
                                loss, pformat(svalues))) from e
            return res

        def fct_grad(coef, x, y):
            'Computes the gradient function for every X and y.'
            svalues = {symbol_N: N}
            for p, c in zip(symbol_params, coef):
                svalues[p] = c

            res = numpy.zeros((len(pnames), ), dtype=x.dtype)
            for n, v in zip(symbol_clq, x):
                svalues[n] = v
            for n, v in zip(symbol_d_clq, y):
                svalues[n] = v
            for row in grads:
                for i, g in enumerate(row):
                    if isinstance(g, Zero):
                        continue
                    try:
                        res[i] = g.evalf(subs=svalues)
                    except (AttributeError, TypeError,  # pragma: no cover
                            IndexError) as e:
                        raise RuntimeError(
                            'Unable to calculate gradient for [{}] with '
                            'values={}.'.format(
                                g, pformat(svalues))) from e
            return res

        coef = numpy.array([self[p] for p in pnames])
        lrn = 1. / (N ** 1.5)
        sgd = SGDOptimizer(
            coef, learning_rate_init=learning_rate_init * lrn,
            lr_schedule=lr_schedule, momentum=momentum,
            power_t=power_t)

        sgd.train(X, y, fct_loss, fct_grad, max_iter=max_iter,
                  verbose=verbose)

        # uses trained coefficients
        coef = sgd.coef
        for n, c in zip(pnames, coef):
            self[n] = c



# coding: utf-8
"""
Implementation of a model for epidemics propagation.
"""
from ._base_sir import BaseSIR


class CovidSIR(BaseSIR):
    """
    Inspiration `Modelling some COVID-19 data
    <http://webpopix.org/covidix19.html>`_.

    .. runpython::
        :showcode:
        :rst:

        from aftercovid.models import CovidSIR

        model = CovidSIR()
        print(model.to_rst())

    .. plot::

        from pandas import DataFrame
        import matplotlib.pyplot as plt
        from aftercovid.models import CovidSIR

        model = CovidSIR()
        sims = list(model.iterate(60))
        df = DataFrame(sims)
        print(df.head())
        ax = df.plot(y=['S', 'I', 'R', 'D'], kind='line')
        ax.set_xlabel("jours")
        ax.set_ylabel("population")
        r0 = model.R0()
        ax.set_title("Simulation SIR\\nR0=%f" % r0)

        plt.show()

    Visual representation:

    .. gdot::
        :script:

        from aftercovid.models import CovidSIR
        model = CovidSIR()
        print(model.to_dot())

    See :ref:`l-base-model-sir` to get the methods
    common to SIRx models.
    """

    P0 = [
        ('beta', 0.5, 'taux de transmission dans la population'),
        ('mu', 1 / 14., '1/. : durée moyenne jusque la guérison'),
        ('nu', 1 / 21., '1/. : durée moyenne jusqu\'au décès'),
    ]

    Q0 = [
        ('S', 9990., 'personnes non contaminés'),
        ('I', 10., 'nombre de personnes malades ou contaminantes'),
        ('R', 0., 'personnes guéries (recovered)'),
        ('D', 0., 'personnes décédées'),
    ]

    C0 = [
        ('N', 10000., 'population'),
    ]

    eq = {
        'S': '- beta * S / N * I',
        'I': 'beta * S / N * I - mu * I - nu * I',
        'R': 'mu * I',
        'D': 'nu * I'
    }

    def __init__(self):
        BaseSIR.__init__(
            self,
            p=CovidSIR.P0.copy(),
            q=CovidSIR.Q0.copy(),
            c=CovidSIR.C0.copy(),
            eq=CovidSIR.eq.copy())

    def R0(self, t=0):
        '''Returns R0 coefficient.'''
        return self['beta'] / (self['nu'] + self['mu'])



# coding: utf-8
"""
Exploration on COVID data.
"""

__version__ = "0.1.0"
__author__ = "Xavier Dupré, ..."


def check(verbose=1):
    """
    Runs a couple of functions to check the module is working.

    :param verbose: 0 to hide the standout output
    :return: list of dictionaries, result of each test
    """
    return []



# coding: utf-8
"""
Common methods about simulation for :epkg:`SIR` models.
"""
import numpy
from sympy import Symbol, diff as sympy_diff


class BaseSIRSimulation:
    """
    Common methods about simulation for :epkg:`SIR` models.
    """

    def eqsign(self, eqname, name):
        """
        Returns the sign of the second derivative for equation
        *eqname* against *name*.

        :param eqname: equation name
        :param name: symbol name
        :return: boolean
        """
        eq = self._eq[eqname]
        df = sympy_diff(eq, Symbol(name))
        ev = self.evalf_eq(df)
        return 1 if ev >= 0 else -1

    def evalf_eq(self, eq, t=0):
        """
        Evaluates an :epkg:`sympy` expression.
        """
        svalues = self._eval_cache()
        svalues[self._syms['t']] = t
        for k, v in zip(self._q, self._val_q):
            svalues[self._syms[k[0]]] = v
        return eq.evalf(subs=svalues)

    def _eval_cache(self):
        values = self.cst_param
        svalues = {self._syms[k]: v for k, v in values.items()}
        return svalues

    def eval_diff(self, t=0):
        """
        Evaluates derivatives.
        Returns a dictionary.
        """
        svalues = self._eval_cache()
        svalues[self._syms['t']] = t
        for k, v in zip(self._q, self._val_q):
            svalues[self._syms[k[0]]] = v

        res = {}
        for k, v in self._eq.items():
            res[k] = v.evalf(subs=svalues)
        return res

    def iterate(self, n=10, t=0, derivatives=False):
        """
        Evalues the quantities for *n* iterations.
        Returns a list of dictionaries.
        If *derivatives* is True, it returns two dictionaries.

        :param n: number of iterations
        :param t: first *t*
        :param derivatives: returns the derivative as well
        :param first: add current values stored in the model
        :return: iterator on dictionaries
        """
        svalues = self._eval_cache()
        svalues[self._syms['t']] = t
        vals = {k[0]: v for k, v in zip(self._q, self._val_q)}
        for i in range(t, t + n):

            for k, v in zip(self._q, self._val_q):
                svalues[self._syms[k[0]]] = v
            diff = {}
            for k, v in self._eq.items():
                diff[k] = v.evalf(subs=svalues)

            fvals = {k: float(v) for k, v in vals.items()}
            if derivatives:
                yield fvals, diff
            else:
                yield fvals

            for k, v in diff.items():
                vals[k] += v
            self.update(**vals)

    def iterate2array(self, n=10, t=0, derivatives=False):
        """
        Evalues the quantities for *n* iterations.
        Returns matrices.

        :param n: number of iterations
        :param t: first *t*
        :param derivatives: returns the derivative as well
        :return: iterator on dictionaries
        """
        clq = self.quantity_names
        pos = {n: i for i, n in enumerate(clq)}
        res = list(self.iterate(n=n, t=t, derivatives=derivatives))
        qu = numpy.zeros((len(res), len(clq)), dtype=numpy.float32)
        if derivatives:
            de = numpy.zeros((len(res), len(clq)), dtype=numpy.float32)
            for i, (r, d) in enumerate(res):
                for j, n in enumerate(pos):
                    qu[i, j] = r.get(n, numpy.nan)
                for j, n in enumerate(pos):
                    de[i, j] = d.get(n, numpy.nan)
            return qu, de
        else:
            for i, r in enumerate(res):
                for j, n in enumerate(pos):
                    qu[i, j] = r.get(n, numpy.nan)
            return qu

    def R0(self, t=0):
        '''Returns R0 coefficient.'''
        raise NotImplementedError()



# coding: utf-8
"""
Helpers for :epkg:`sympy`.
"""


def enumerate_traverse(e, parent=None):
    """
    Enumerates all nodes in *e*.

    :param e: :epkg:`sympy` expression or node
    :param parent: parent of *e*
    :return: iterator on itself and the children
    """
    yield dict(e=e, p=parent)
    for arg in e.args:
        for r in enumerate_traverse(arg, e):
            yield r


class SympyNode:
    """
    Model a :epkg:`sympy` expression.
    It probably exist in :epkg:`sympy`.

    :param e: :epkg:`sympy` expression
    :param parent: *SympyNode*
    """

    def __init__(self, e, parent=None):
        if parent is not None and not isinstance(parent, SympyNode):
            raise TypeError(  # pragma: no cover
                "parent must be None or SympyNode")
        self._e = e
        self._parent = parent
        self._children = []
        for a in self._e.args:
            self._children.append(SympyNode(a, self))

    def __repr__(self):
        return 'SympyNode("%r")' % self.element

    @property
    def element(self):
        return self._e

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        return self._children

    def __iter__(self):
        """
        Enumerate all nodes.
        """
        yield self.element
        for a in self.children:
            for n in a:
                yield a

    def enumerate_parents(self):
        """
        Enumerates all parents including itself.
        """
        n = self
        while n.parent is not None:
            yield n
            n = n.parent



# coding: utf-8
"""
Common functions for :epkg:`SIR` models.
"""
import numpy
from sympy import symbols, Symbol
import sympy.printing as printing
from sympy.parsing.sympy_parser import (
    parse_expr, standard_transformations, implicit_application)
from ._sympy_helper import enumerate_traverse
from ._base_sir_sim import BaseSIRSimulation
from ._base_sir_estimation import BaseSIRSklAPI


class BaseSIR(BaseSIRSimulation, BaseSIRSklAPI):
    """
    Base model for :epkg:`SIR` models.

    :param p: list of `[(name, initial value or None, comment)]` (parameters)
    :param q: list of `[(name, initial value or None, comment)]` (quantities)
    :param c: list of `[(name, initial value or None, comment)]` (constants)
    :param eq: equations
    """

    def __init__(self, p, q, c=None, eq=None):
        if not isinstance(p, list):
            raise TypeError("p must be a list of tuple.")
        if not isinstance(q, list):
            raise TypeError("q must be a list of tuple.")
        if not isinstance(c, list):
            raise TypeError("c must be a list of tuple.")
        if eq is not None and not isinstance(eq, dict):
            raise TypeError("eq must be a dictionary.")
        self._p = p
        self._q = q
        self._c = c
        if eq is not None:
            locs = {'t': symbols('t', cls=Symbol)}
            for v in self._p:
                locs[v[0]] = symbols(v[0], cls=Symbol)
            for v in self._c:
                locs[v[0]] = symbols(v[0], cls=Symbol)
            for v in self._q:
                locs[v[0]] = symbols(v[0], cls=Symbol)
            self._syms = locs
            tr = standard_transformations + (implicit_application, )
            self._eq = {}
            for k, v in eq.items():
                try:
                    self._eq[k] = parse_expr(v, locs, transformations=tr)
                except (TypeError, ValueError) as e:  # pragma: no cover
                    raise RuntimeError(
                        "Unable to parse '{}'.".format(v)) from e
        else:
            self._eq = None
        self._init()

    def _init(self):
        """
        Starts from the initial values.
        """
        def _def_(name, v):
            if v is not None:
                return v
            if name == 'N':
                return 10000.
            return 0.

        self._val_p = numpy.array([_def_(v[0], v[1]) for v in self._p])
        self._val_q = numpy.array([_def_(v[0], v[1]) for v in self._q])
        self._val_c = numpy.array([_def_(v[0], v[1]) for v in self._c])

    def get_index(self, name):
        '''
        Returns the index of a name (True or False, position).
        '''
        for i, v in enumerate(self._p):
            if v[0] == name:
                return 'p', i
        for i, v in enumerate(self._q):
            if v[0] == name:
                return 'q', i
        for i, v in enumerate(self._c):
            if v[0] == name:
                return 'c', i
        raise ValueError("Unable to find name '{}'.".format(name))

    def __setitem__(self, name, value):
        """
        Updates a value whether it is a parameter or a quantity.

        :param name: name
        :param value: new value
        """
        p, pos = self.get_index(name)
        if p == 'p':
            self._val_p[pos] = value
        elif p == 'q':
            self._val_q[pos] = value
        elif p == 'c':
            self._val_c[pos] = value

    def __getitem__(self, name):
        """
        Retrieves a value whether it is a parameter or a quantity.

        :param name: name
        :return: value
        """
        p, pos = self.get_index(name)
        if p == 'p':
            return self._val_p[pos]
        if p == 'q':
            return self._val_q[pos]
        if p == 'c':
            return self._val_c[pos]

    @property
    def names(self):
        'Returns the list of names.'
        return list(sorted(
            [v[0] for v in self._p] + [v[0] for v in self._q] +
            [v[0] for v in self._c]))

    @property
    def quantity_names(self):
        'Returns the list of quantities names (unsorted).'
        return [v[0] for v in self._q]

    @property
    def param_names(self):
        'Returns the list of parameters names (unsorted).'
        return [v[0] for v in self._p]

    @property
    def P(self):
        '''
        Returns the parameters
        '''
        return self._p

    @property
    def Q(self):
        '''
        Returns the quantities
        '''
        return self._q

    @property
    def C(self):
        '''
        Returns the quantities
        '''
        return self._c

    def update(self, **values):
        """Updates values."""
        for k, v in values.items():
            self[k] = v

    def get(self):
        """Retrieves all values."""
        return {n: self[n] for n in self.names}

    def to_rst(self):
        '''
        Returns a string formatted in RST.
        '''
        rows = [
            '*{}*'.format(self.__class__.__name__),
            '',
            '*Q*',
            ''
        ]
        for name, _, doc in self._q:
            rows.append('* *{}*: {}'.format(name, doc))
        rows.extend(['', '*C*', ''])
        for name, _, doc in self._c:
            rows.append('* *{}*: {}'.format(name, doc))
        rows.extend(['', '*P*', ''])
        for name, _, doc in self._p:
            rows.append('* *{}*: {}'.format(name, doc))
        if self._eq is not None:
            rows.extend(['', '*E*', '', '.. math::',
                         '', '    \\begin{array}{l}'])
            for i, (k, v) in enumerate(sorted(self._eq.items())):
                line = "".join(
                    ["    ", "\\frac{d%s}{dt} = " % k, printing.latex(v)])
                if i < len(self._eq) - 1:
                    line += " \\\\"
                rows.append(line)
            rows.append("    \\end{array}")

        return '\n'.join(rows)

    def enumerate_edges(self):
        """
        Enumerates the list of quantities contributing
        to others. It ignores constants.
        """
        if self._eq is not None:
            params = set(_[0] for _ in self.P)
            quants = set(_[0] for _ in self.Q)
            for k, v in sorted(self._eq.items()):
                n2 = k
                for dobj in enumerate_traverse(v):
                    term = dobj['e']
                    if not hasattr(term, 'name'):
                        continue
                    if term.name not in params:
                        continue
                    parent = dobj['p']
                    others = list(
                        _['e'] for _ in enumerate_traverse(parent))
                    for o in others:
                        if hasattr(o, 'name') and o.name in quants:
                            sign = self.eqsign(n2, o.name)
                            yield (sign, o.name, n2, term.name)

    def to_dot(self, verbose=False, full=False):
        """
        Produces a graph in :epkg:`DOT` format.
        """
        rows = ['digraph{']

        pattern = ('    {name} [label="{name}\\n{doc}" shape=record];'
                   if verbose else
                   '    {name} [label="{name}"];')
        for name, _, doc in self._q:
            rows.append(pattern.format(name=name, doc=doc))
        for name, _, doc in self._c:
            rows.append(pattern.format(name=name, doc=doc))

        if self._eq is not None:
            pattern = (
                '    {n1} -> {n2} [label="{sg}{name}\\nvalue={v:1.2g}"];'
                if verbose else '    {n1} -> {n2} [label="{sg}{name}"];')
            for sg, a, b, name in set(self.enumerate_edges()):
                if not full and (a == b or sg < 0):
                    continue
                value = self[name]
                stsg = '' if sg > 0 else '-'
                rows.append(
                    pattern.format(n1=a, n2=b, name=name, v=value, sg=stsg))

        rows.append('}')
        return '\n'.join(rows)

    @property
    def cst_param(self):
        '''
        Returns a dictionary with the constant and the parameters.
        '''
        res = {}
        for k, v in zip(self._c, self._val_c):
            res[k[0]] = v
        for k, v in zip(self._p, self._val_p):
            res[k[0]] = v
        return res



# -*- coding: utf-8 -*-
import os
import sys
from distutils.command.build_ext import build_ext
from setuptools import setup, Extension
from setuptools import find_packages
import numpy

##########
# helpers
##########


# Placeholder to change the default
# compilation for command ``build_ext --inplace``.
def get_cmd_classes():

    class build_ext_subclass(build_ext):
        def build_extensions(self):
            build_ext.build_extensions(self)

    return {'build_ext': build_ext_subclass}


def get_defined_args():
    if sys.platform.startswith("win"):
        # windows
        define_macros = [('USE_OPENMP', None)]
        libraries = ['kernel32']
        extra_compile_args = ['/EHsc', '/O2', '/Gy', '/openmp']
        extra_link_args = None
    elif sys.platform.startswith("darwin"):
        # mac osx
        define_macros = [('USE_OPENMP', None)]
        libraries = None
        extra_compile_args = ['-lpthread', '-stdlib=libc++',
                              '-mmacosx-version-min=10.7', '-Xpreprocessor',
                              '-fopenmp']
        extra_link_args = ["-lomp"]
    else:
        # linux
        define_macros = [('USE_OPENMP', None)]
        libraries = None
        extra_compile_args = ['-lpthread', '-fopenmp']
        # option '-mavx2' forces the compiler to use
        # AVX instructions the processor might not have
        extra_link_args = ['-lgomp']

    return {
        'define_macros': define_macros,
        'libraries': libraries,
        'extra_compile_args': extra_compile_args,
        'extra_link_args': extra_link_args,
    }


def get_extension_tutorial(name):
    pattern1 = "aftercovid.tutorial.%s"
    srcs = ['aftercovid/tutorial/%s.pyx' % name]
    args = get_defined_args()
    if name in ['dot_cython', 'experiment_cython', 'dot_cython_omp']:
        srcs.extend(['aftercovid/tutorial/%s_.cpp' % name])
        args['language'] = "c++"

    ext = Extension(pattern1 % name, srcs,
                    include_dirs=[numpy.get_include()],
                    **args)

    opts = dict(boundscheck=False, cdivision=True,
                wraparound=False, language_level=3,
                cdivision_warnings=True)

    ext_modules = []

    from Cython.Build import cythonize
    ext_modules.extend(cythonize([ext], compiler_directives=opts))
    return ext_modules


######################
# beginning of setup
######################


here = os.path.dirname(__file__)
if here == "":
    here = '.'
packages = find_packages(where=here)
package_dir = {k: os.path.join(here, k.replace(".", "/")) for k in packages}
package_data = {
    "aftercovid.tutorial": ["*.pyx", '*.cpp', '*.h'],
}

try:
    with open(os.path.join(here, "requirements.txt"), "r") as f:
        requirements = f.read().strip(' \n\r\t').split('\n')
except FileNotFoundError:
    requirements = []
if len(requirements) == 0 or requirements == ['']:
    requirements = []

try:
    with open(os.path.join(here, "readme.rst"), "r", encoding='utf-8') as f:
        long_description = "aftercovid:" + f.read().split('aftercovid:')[1]
except FileNotFoundError:
    long_description = ""

version_str = '0.0.1'
with open(os.path.join(here, 'aftercovid/__init__.py'), "r") as f:
    line = [_ for _ in [_.strip("\r\n ")
                        for _ in f.readlines()]
            if _.startswith("__version__")]
    if len(line) > 0:
        version_str = line[0].split('=')[1].strip('" ')

requires = ["cython", "fire", "numpy", "pandas", "scipy", "sympy"]
with open(os.path.join(here, 'requirements.txt'), "r") as f:
    requires = [_.strip() for _ in f.readlines()]
    requires = [_ for _ in requires if _]

ext_modules = []
for ext in ['experiment_cython']:
    ext_modules.extend(get_extension_tutorial(ext))

setup(name='aftercovid',
      version=version_str,
      description="Code, ideas about COVID epidemics",
      long_description=long_description,
      author='Xavier Dupré, ...',
      author_email='xavier.dupre@gmail.com',
      url='https://github.com/sdpython/aftercovid',
      ext_modules=ext_modules,
      packages=packages,
      package_dir=package_dir,
      package_data=package_data,
      setup_requires=requires,
      install_requires=requires,
      cmdclass=get_cmd_classes())



"""
Implements simple stochastic gradient optimisation.
It is inspired from `_stochastic_optimizers.py
<https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/
neural_network/_stochastic_optimizers.py>`_.
"""
import numpy


class BaseOptimizer:
    """
    Base stochastic gradient descent optimizer.

    Parameters
    ----------

    coef : array, initial coefficient

    learning_rate_init : float, default=0.1
        The initial learning rate used. It controls the step-size in updating
        the weights

    Attributes
    ----------
    learning_rate : float
        the current learning rate
    """

    def __init__(self, coef, learning_rate_init=0.1):
        if not isinstance(coef, numpy.ndarray):
            raise TypeError("coef must be an array.")
        self.coef = coef
        self.learning_rate_init = learning_rate_init
        self.learning_rate = float(learning_rate_init)

    def _get_updates(self, grad):
        raise NotImplementedError("Must be overwritten.")  # pragma no cover

    def update_coef(self, grad):
        """
        Updates coefficients with given gradient.

        :param grad: array, gradient
        """
        if self.coef.shape != grad.shape:
            raise ValueError("coef and grad must have the same shape.")
        update = self._get_updates(grad)
        self.coef += update

    def iteration_ends(self, time_step):
        """
        Performs update to learning rate and potentially other states at the
        end of an iteration.
        """
        pass  # pragma: no cover

    def train(self, X, y, fct_loss, fct_grad, max_iter=100, verbose=False):
        """
        Optimizes the coefficients.

        :param X: datasets (array)
        :param y: expected target
        :param fct_loss: loss function, signature: `f(coef, X, y) -> float`
        :param fct_grad: gradient function, signature: `g(coef, x, y) -> array`
        :param max_iter: number maximum of iteration
        :param verbose: display information
        :return: loss
        """
        if not isinstance(X, numpy.ndarray):
            raise TypeError("X must be an array.")
        if not isinstance(y, numpy.ndarray):
            raise TypeError("y must be an array.")
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of rows.")

        loss = fct_loss(self.coef, X, y)
        if verbose:
            self._display_progress(0, max_iter, loss)
        n_samples = 0
        for it in range(max_iter):
            irows = numpy.random.choice(X.shape[0], X.shape[0])
            for irow in irows:
                grad = fct_grad(self.coef, X[irow, :], y[irow])
                self.update_coef(grad)
                n_samples += 1

            self.iteration_ends(n_samples)
            loss = fct_loss(self.coef, X, y)
            if verbose:
                self._display_progress(it + 1, max_iter, loss)
        return loss

    def _display_progress(self, it, max_iter, loss):
        'Displays training progress.'
        print('{}/{}: loss: {:1.4g}'.format(it, max_iter, loss))


class SGDOptimizer(BaseOptimizer):
    """
    Stochastic gradient descent optimizer with momentum.

    Parameters
    ----------

    coef : array, initial coefficient

    learning_rate_init : float, default=0.1
        The initial learning rate used. It controls the step-size in updating
        the weights

    lr_schedule : {'constant', 'adaptive', 'invscaling'}, default='constant'
        Learning rate schedule for weight updates.
        -'constant', is a constant learning rate given by
         'learning_rate_init'.
        -'invscaling' gradually decreases the learning rate 'learning_rate_'
          at each time step 't' using an inverse scaling exponent of
          'power_t'. learning_rate_ = learning_rate_init / pow(t, power_t)
        -'adaptive', keeps the learning rate constant to
         'learning_rate_init' as long as the training keeps decreasing.
         Each time 2 consecutive epochs fail to decrease the training loss by
         tol, or fail to increase validation score by tol if 'early_stopping'
         is on, the current learning rate is divided by 5.

    momentum : float, default=0.9
        Value of momentum used, must be larger than or equal to 0

    Attributes
    ----------

    learning_rate : float the current learning rate

    velocity : array, velocity that are used to update params

    Example
    -------

    The following example how to optimize a simple linear regression.

    .. runpython::
        :showcode:

        import numpy
        from aftercovid.optim import SGDOptimizer


        def fct_loss(c, X, y):
            return numpy.linalg.norm(X @ c - y) ** 2


        def fct_grad(c, x, y):
            return x * (x @ c - y) * 0.1


        coef = numpy.array([0.5, 0.6, -0.7])
        X = numpy.random.randn(10, 3)
        y = X @ coef

        sgd = SGDOptimizer(numpy.random.randn(3))
        sgd.train(X, y, fct_loss, fct_grad, max_iter=15, verbose=True)
        print('optimized coefficients:', sgd.coef)
    """

    def __init__(self, coef, learning_rate_init=0.1, lr_schedule='constant',
                 momentum=0.9, power_t=0.5):
        super().__init__(coef, learning_rate_init)
        self.lr_schedule = lr_schedule
        self.momentum = momentum
        self.power_t = power_t
        self.velocity = numpy.zeros_like(coef)

    def iteration_ends(self, time_step):
        """
        Performs updates to learning rate and potential other states at the
        end of an iteration.

        :param time_step: int
            number of training samples trained on so far, used to update
            learning rate for 'invscaling'
        """
        if self.lr_schedule == 'invscaling':
            self.learning_rate = (float(self.learning_rate_init) /
                                  (time_step + 1) ** self.power_t)

    def _get_updates(self, grad):
        """
        Gets the values used to update params with given gradients.

        :param grad: array, gradient
        :return: updates, array, the values to add to params
        """
        update = self.momentum * self.velocity - self.learning_rate * grad
        self.velocity = update
        return update

    def _display_progress(self, it, max_iter, loss):
        'Displays training progress.'
        print('{}/{}: loss: {:1.4g} lr={:1.3g}'.format(
            it, max_iter, loss, self.learning_rate))
